import SensorJSDriver from '../SensorJSDriver';
import * as SensorJSInterface from '../SensorJSInterface';
import log from 'loglevel';

import * as Ably from 'ably/build/ably.js';

export default class AblyRealtime extends SensorJSDriver {

    constructor(options, callback) {
        super('ably');

        const channels = [];

        this.connect = ()=> {
            log.info('ably:: connecting')
            this.client = new Ably.Realtime(options);

            this.client.connection.once('connected', () => {
                log.info('ably::connected');
                callback?.(this);
            });

            // indefinite/permanent failure - missing or wrong credentials
            this.client.connection.on('failed', () => {
                log.error('ably::connection_failed');
                callback?.(this, 'failed');
            });

            this.client.connection.on('disconnected', () => {
                log.warn('ably::connection_disconnected');
                callback?.(this, 'disconnected');
            });

            this.client.connection.on('suspended', () => {
                log.warn('ably::connection_suspended');
                callback?.(this, 'suspended');
            });

            // connection closed by the user
            this.client.connection.on('closed', () => {
                log.info('ably::connection_closed');
                callback?.(this, 'closed');
            });
        }

        this.reconnect = ()=> {
            this.client?.connection.connect();
            return this;
        }

        this.disconnect = ()=> {
            this.client?.connection.close();
        }

        this.release = ()=> {
            log.debug('ably::release');
            channels.forEach((channel)=>channel.off(this.handleMessages));
            channels.length = 0;
            this.client?.connection.close();
            this.listeners.length = 0;
        }

        this.handleMessages = (message) => {
            log.debug('ably::received message: ', message);
            this.listeners.forEach((listener) => {
                listener(message.data.status, message.data);
            });
        }

        this.subscribe = (channelId, sid, messageListener)=> {
            log.debug('ably::subscribed to channel: ', channelId);
            const channel = this.client.channels.get(channelId);
            if (!channel) {
                log.error('ably:: unable to retrieve a channel with id: ', channelId);
                if (callback) callback(this, 'challen_creation_failed');
            }
            channels[channelId] = channel;
            if (sid) {
                log.debug('ably::with sid: ', sid);
                log.debug('ably::and listener: ', messageListener);    
                this._listeners[sid] = messageListener;
                channel.subscribe(messageListener);
            } else {
                channel.subscribe(this.handleMessages);
            }
            return this;
        }

        this.unsubscribe = (channelId, sid) => {
            log.debug('ably::unsubscribed from channel: ', channelId);
            const channel = channels[channelId];
            if (!channel) {
                log.error('ably:: unable to retrieve a channel with id: ', channelId);
                if (callback) callback(this, 'challen_creation_failed');
            }
            if (sid) {
                log.debug('ably::with sid: ', sid);
                log.debug('ably::and listener: ', this._listeners[sid]);
                channel.unsubscribe(this._listeners[sid]);
                delete this._listeners[sid];
            } else {
                channel.unsubscribe(this.handleMessages);
            }
            delete channels[channelId];
            return this;
        }

        this.publish = (channelId, name, message, callback) => {
            log.debug('ably::publishing message ', JSON.stringify(message));
            const channel = channels[channelId];
            if (!channel) {
                log.error('ably:: unable to retrieve a channel with id: ', channelId);
                if (callback) callback(this, 'challen_creation_failed');
            }
            channel.publish(name, message, callback);
            return this;
        }

        this.presence = (channelId) => channels[channelId].presence;

        this.connected = () => this.client?.connection?.state === 'connected';

        this.connection = () => this.client?.connection;

        // check interface implementation
        SensorJSInterface.check(this);

        // start connection
        this.connect();

        return this;
    }

}